# oops-repository is Copyright 2011 Canonical Ltd.
#
# Canonical Ltd ("Canonical") distributes the oops-repository source code under
# the GNU Affero General Public License, version 3 ("AGPLv3"). See the file
# LICENSE in the source tree for more information.

"""basic operations on oopses in the db."""

import json
import time
import uuid

import pycassa
from pycassa.cassandra.ttypes import NotFoundException

DAY = 60*60*24
MONTH = DAY*30

_connection_pool = None
def connection_pool(config):
    '''Caching constructor of pycassa.ConnectionPool'''
    global _connection_pool
    if _connection_pool:
        return _connection_pool
    # I have no idea why max_retries is evaluating as 0 when not set, but here
    # we are, brute forcing this.
    creds = {'username': config['username'], 'password': config['password']}
    _connection_pool = pycassa.ConnectionPool(config['keyspace'],
                                              config['host'],
                                              max_retries=5,
                                              credentials=creds)
    return _connection_pool

def prune(config):
    """Remove OOPSES that are over 30 days old."""
    pool = connection_pool(config)
    dayoops_cf = pycassa.ColumnFamily(pool, 'DayOOPS')
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')
    # Find days to prune
    days = set()
    prune_to = time.strftime('%Y%m%d', time.gmtime(time.time() - MONTH))
    for key, _ in dayoops_cf.get_range():
        if key < prune_to:
            days.add(key)
    if not days:
        return
    # collect all the oopses (buffers all in memory; may want to make
    # incremental in future)
    batch_size = 10000
    oopses = []
    for day in days:
        columns_found = batch_size
        start_col = ''
        while columns_found==batch_size:
            columns = dayoops_cf.get(day, column_start=start_col)
            columns_found = len(columns)
            for column, oopsid in columns.items():
                start_col = column
                oopses.append(oopsid)
    # Remove the oopses
    batch = oops_cf.batch()
    map(batch.remove, oopses)
    batch.send()
    # Clean out the days aggregates
    # Clean out the days index
    batch = dayoops_cf.batch()
    map(batch.remove, days)
    batch.send()

def insert(config, oopsid, oops_json, user_token=None, fields=None):
    """Insert an OOPS into the system.
    
    :return: The day which the oops was filed under.
    """
    # make sure the oops report is a json dict, and break out each key to a
    # separate column. For now, rather than worrying about typed column values
    # we just coerce them all to strings.
    oops_dict = json.loads(oops_json)
    assert isinstance(oops_dict, dict)
    insert_dict = {}
    for key, value in oops_dict.items():
        insert_dict[key] = json.dumps(value)
    return _insert(config, oopsid, insert_dict, user_token, fields)

def insert_dict(config, oopsid, oops_dict, user_token=None, fields=None):
    """Insert an OOPS into the system.
    
    :return: The day which the oops was filed under.
    """
    assert isinstance(oops_dict, dict)
    return _insert(config, oopsid, oops_dict, user_token, fields)

def _insert(config, oopsid, insert_dict, user_token=None, fields=None):
    """Internal function. Do not call this directly.

    :param oopsid: The identifier for this OOPS.
    :param insert_dict: A dictionary containing the data to associate this OOPS
        with.
    :param user_token: An identifier for the user who experienced this OOPS.
    :param fields: A dictionary containing keys to increment counters for.
    :return: The day which the oops was filed under.
    """
    pool = connection_pool(config)
    dayoops_cf = pycassa.ColumnFamily(pool, 'DayOOPS')
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')
    useroops_cf = pycassa.ColumnFamily(pool, 'UserOOPS')
    dayusers_cf = pycassa.ColumnFamily(pool, 'DayUsers')
    # Retry counter increments when a timeout is encountered. This has the risk
    # of double-counting, but the only alternative is to under count.
    counters_cf = pycassa.ColumnFamily(pool, 'Counters',
                                       retry_counter_mutations=True)
    day_key = time.strftime('%Y%m%d', time.gmtime())
    now_uuid = uuid.uuid1()

    oops_cf.insert(oopsid, insert_dict)
    dayoops_cf.insert(day_key, {now_uuid:oopsid})
    # Provide quick lookups of the total number of oopses for the day by
    # maintaining a counter.
    counters_cf.add('oopses', day_key)
    if fields:
        for field in fields:
            counters_cf.add('oopses:%s' % field, day_key)

    if user_token:
        useroops_cf.insert(user_token, {oopsid : ''})
        # This along with DayOOPS lets us calculate the average number of
        # crashes for the day (total crashes / total users).
        try:
            dayusers_cf.get(day_key, [user_token])
        except NotFoundException:
            # Provide quick lookups of the total number of users for the day by
            # maintaining a counter. Only increment the counter if we have not
            # seen this user today.
            counters_cf.add('users', day_key)
            if fields:
                for field in fields:
                    counters_cf.add('users:%s' % field, day_key)
            dayusers_cf.insert(day_key, {user_token : ''})

        if fields:
            for field in fields:
                dayusers_cf.insert('%s:%s' % (field, day_key), {user_token : ''})

    return day_key

def bucket(config, oopsid, bucketid, fields=None):
    """Adds an OOPS to a bucket, a collection of OOPSes that form a single
    issue. If the bucket does not exist, it will be created.
    
    :return: The day which the bucket was filed under.
    """
    pool = connection_pool(config)
    bucket_cf = pycassa.ColumnFamily(pool, 'Bucket')
    daybucket_cf = pycassa.ColumnFamily(pool, 'DayBuckets')
    daybucketcount_cf = pycassa.ColumnFamily(pool, 'DayBucketsCount',
                                             retry_counter_mutations=True)
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')

    # Get the timestamp
    timestamp = oops_cf.get(oopsid, column_count=1, include_timestamp=True)
    timestamp = timestamp.values()[0][1]
    day_key = time.strftime('%Y%m%d', time.gmtime(timestamp / 1000000))

    bucket_cf.insert(bucketid, {pycassa.util.uuid.UUID(oopsid) : ''})
    daybucket_cf.insert((day_key, bucketid), {oopsid : ''})
    
    if fields is not None:
        resolutions = (day_key[:4], day_key[:6], day_key)
        # All buckets for the given resolution.
        for field in fields:
            for resolution in resolutions:
                # We have no way of knowing whether an increment has been
                # performed if the write fails unexpectedly (CASSANDRA-2495).
                # We will apply eventual consistency to this problem and
                # tolerate slightly inaccurate counts for the span of a single
                # day, cleaning up once this period has passed. This will be
                # done by counting the number of columns in DayBuckets for the
                # day and bucket ID.
                daybucketcount_cf.add(':'.join((field, resolution)), bucketid)
        for resolution in resolutions:
            daybucketcount_cf.add(resolution, bucketid)
    return day_key

def update_bucket_versions(config, bucketid, version):
    pool = connection_pool(config)
    bucketversions_cf = pycassa.ColumnFamily(pool, 'BucketVersions',
                                             retry_counter_mutations=True)
    bucketversions_cf.add(bucketid, version)

def query_bucket_versions(config, bucketid):
    pool = connection_pool(config)
    bucketversions_cf = pycassa.ColumnFamily(pool, 'BucketVersions')
    try:
        return bucketversions_cf.xget(bucketid)
    except NotFoundException:
        return None 

def update_bucket_metadata(config, bucketid, source, version, comparator, release=''):
    # We only update the first and last seen version fields. We do not update
    # the current version field as talking to Launchpad is an expensive
    # operation, and we can do that out of band.
    pool = connection_pool(config)
    bucketmetadata_cf = pycassa.ColumnFamily(pool, 'BucketMetadata')
    metadata = pycassa.util.OrderedDict()
    try:
        metadata = bucketmetadata_cf.get(bucketid)
        lastseen = metadata.get('LastSeen', None)
        if (not lastseen or comparator(lastseen, version) < 0):
            metadata['LastSeen'] = version
        firstseen = metadata.get('FirstSeen', None)
        if (not firstseen or comparator(firstseen, version) > 0):
            metadata['FirstSeen'] = version
    except NotFoundException:
        metadata['FirstSeen'] = version
        metadata['LastSeen'] = version

    if release:
        k = '~%s:FirstSeen' % release
        firstseen = metadata.get(k, None)
        if (not firstseen or comparator(firstseen, version) > 0):
            metadata[k] = version
        k = '~%s:LastSeen' % release
        lastseen = metadata.get(k, None)
        if (not lastseen or comparator(lastseen, version) < 0):
            metadata[k] = version

    if metadata:
        metadata['Source'] = source
        bucketmetadata_cf.insert(bucketid, metadata)

def update_bucket_systems(config, bucket, system):
    if not system:
        return
    pool = connection_pool(config)
    bucketsystems_cf = pycassa.ColumnFamily(pool, 'BucketSystems')
    bucketsystems_cf.insert(bucket, {system: ''})
